from apicrud import BasicCRUD
import apicrud.geocode as geocode

fields = ('address', 'neighborhood', 'city', 'state', 'country')


class LocationController(BasicCRUD):

    def __init__(self):
        super().__init__(resource='location')

    @staticmethod
    def create(body):
        """invoke geocoder to store coordinates
        """
        if body.get('city') and body.get('state'):
            body['geolat'], body['geolong'], body['neighborhood'] = (
                geocode.lookup(**{item: value for item, value
                                  in body.items() if item in fields}))
        return super(LocationController, LocationController).create(body)

    @staticmethod
    def update(id, body):
        """invoke geocoder to store coordinates
        """
        if body.get('city') and body.get('state'):
            if body.get('neighborhood') and (
                    body['neighborhood'][:14] == 'A neighborhood'):
                body['neighborhood'] = None
            body['geolat'], body['geolong'], body['neighborhood'] = (
                geocode.lookup(**{item: value for item, value
                                  in body.items() if item in fields}))
        return super(LocationController, LocationController).update(id, body)

    @staticmethod
    def find(**kwargs):
        """after fetching records, convert geocoded values to
        [longitude, latitude]
        """
        results, status = super(LocationController, LocationController).find(
            **kwargs)
        for record in results['items']:
            if 'geolat' in record:
                access = 'r' if record['privacy'] == 'public' else None
                geocode.with_privacy(record, access)
        return results, status



"""storage.py

Storage access
  Files, photos and videos are stored as objects in a cloud vendor's
  storage service (Amazon S3 or equivalent). The backend server here
  acts as a broker for authenticating uploads and storing/retrieving
  metadata descriptors for each object. Metadata is stored in a
  redis cache.

created 25-jan-2020 by richb@instantlinux.net
"""

from b2sdk.v1 import B2Api, InMemoryAccountInfo
import base64
import boto3
from botocore.config import Config
from botocore.exceptions import ClientError
from datetime import datetime
from flask import g
from flask_babel import _
import json
import logging
import re
import redis
from sqlalchemy.orm.exc import NoResultFound
import uuid

from ..access import AccessControl
from ..const import Constants
from ..grants import Grants
from ..metrics import Metrics
from ..service_config import ServiceConfig
import apicrud.utils as utils

saved_redis = None


class StorageAPI(object):
    """Storage API

    Attributes:
      credential_ttl (int): how long until temp credential expires
      redis_conn (obj): connection to redis
      uid (str): User ID
      db_session (obj): existing db session
    """

    def __init__(self, credential_ttl=None, redis_conn=None,
                 uid=None, db_session=None):
        global saved_redis

        config = ServiceConfig().config
        self.models = ServiceConfig().models
        self.redis_conn = (
            redis_conn or saved_redis or redis.Redis(
                host=config.REDIS_HOST, port=config.REDIS_PORT, db=0))
        saved_redis = self.redis_conn
        self.cache_ttl = config.REDIS_TTL
        self.credential_ttl = credential_ttl or 86400
        self.uid = uid or AccessControl().uid
        self.db = db_session or g.db

    def get_upload_url(self, body):
        """Get a new pre-authorized URL from the storage vendor which
        allows the end user to upload one object to storage

        Args:
          body (dict):
            storage_id key provides ID in storage model; other keys provide
            metadata about the object to be uploaded, which will be stored
            in the new object's entry in database upon upload completion
        """
        storage_id = body.get('storage_id')
        try:
            storage = self.db.query(self.models.Storage).filter_by(
                id=storage_id, status='active').one()
        except NoResultFound:
            return dict(id=storage_id, message='storage volume not found'), 404
        resource = body.get('parent_type', 'album')
        if re.compile('^[a-f0-9]{8}-?[a-f0-9]{4}-?4[a-f0-9]{3}'
                      '-?[89ab][a-f0-9]{3}-?[a-f0-9]{12}.jpeg').match(
                          body.get('name').lower()):
            # Make iPhone filename shorter
            body['name'] = base64.urlsafe_b64encode(
                uuid.UUID(body['name'].strip('.jpeg')).bytes).rstrip(
                    b'=').decode('ascii') + '.jpeg'
        logmsg = dict(uid=self.uid, resource=resource, name=body.get('name'),
                      size=body.get('size'))

        model = dict(album=self.models.Album, list=self.models.List,
                     message=self.models.Message)[resource]
        if resource == 'album':
            try:
                record = self.db.query(model).filter_by(
                    id=body.get('parent_id'), status='active')
                album_size = len(record.one().pictures)
            except NoResultFound:
                return dict(album_id=body.get(
                    'parent_id'), message='album not found'), 404
            max_size = Grants(db_session=self.db).get('album_size')
            if album_size >= max_size:
                msg = _('album is full (max=%d)') % max_size
                logging.info(dict(message=msg, **logmsg))
                return dict(message=msg), 405
        if not AccessControl(model=model).with_permission(
                    'u', query=record):
            msg = _('access_denied')
            logging.warning(dict(message=msg, **logmsg))
            return dict(message=msg), 403

        self.vendor = storage.credentials.vendor
        id = utils.gen_id(prefix='f-')
        if body.get('content_type'):
            ctype = body['content_type'].split('/')[1]
        else:
            ctype = None
        storage_path = ('%s/%s/%s' % (
            storage.prefix if storage.prefix else '', self.uid, id)).strip('/')
        suffix = '.' + ctype if ctype else ''
        if ctype in Constants.MIME_VIDEO_TYPES:
            duration_max = Grants(
                db_session=self.db).get('video_duration_max')
            if not Metrics(uid=self.uid, db_session=self.db).check(
                    'video_daily_total'):
                msg = _(u'daily video upload limit exceeded')
                logging.warning(dict(message=msg, **logmsg))
                return dict(message=msg), 405
            if body.get('duration') and body['duration'] > duration_max:
                msg = _('video exceeds maximum duration=%f') % duration_max
                logging.warning(dict(message=msg, **logmsg))
                return dict(message=msg), 405
        elif ctype in Constants.MIME_IMAGE_TYPES:
            if not Metrics(uid=self.uid, db_session=self.db).check(
                    'photo_daily_total'):
                msg = _(u'daily photo upload limit exceeded')
                logging.warning(dict(message=msg, **logmsg))
                return dict(message=msg), 405
        max_size = Grants(
            db_session=self.db).get('media_size_max')
        if body.get('size') and body['size'] > max_size:
            msg = _('file size exceeds max=%d') % max_size
            logging.info(dict(message=msg, **logmsg))
            return dict(message=msg), 405
        if self.vendor == 'backblaze':
            result, status = StorageBackblaze(
                self.models.Storage).get_upload_url(
                    storage_id, storage_path + suffix,
                    content_type=body.get('content_type'))
        elif self.vendor == 'aws':
            result, status = StorageS3(
                credentials=storage.credentials,
                credential_ttl=self.credential_ttl).get_upload_url(
                    storage.bucket, storage_path + suffix,
                    content_type=body.get('content_type'))
        else:
            return dict(message='Unsupported vendor=%s' % self.vendor), 405
        if status != 201:
            return result, status
        meta = dict(
            created=g.request_start_time.timestamp(),
            ctype=ctype,
            duration=body.get('duration'),
            fid=id,
            height=body.get('height'),
            modified=body.get('modified'),
            name=body.get('name'),
            oid=storage.uid,
            path=storage_path,
            pid=body.get('parent_id'),
            sid=storage_id,
            size=body.get('size'),
            uid=self.uid,
            width=body.get('width'))
        self.update_file_meta(id, meta)
        logging.info(dict(action='get_upload_url', **logmsg))
        return dict(params=result['fields'], upload_url=result['uploadUrl'],
                    file_id=id), status

    def get_file_meta(self, file_id):
        """Retrieve metadata from redis cache

        Args:
          file_id (str): ID in file model
        Returns:
          dict: fid, uid, sid, pid, path, name
        """
        key = 'fid:%s:%s' % (self.uid, file_id[-8:])
        try:
            meta = json.loads(self.redis_conn.get(key))
        except TypeError:
            return None
        except Exception as ex:
            logging.warning('action=get_upload_state uid=%s exception=%s' %
                            (self.uid, str(ex)))
            return None
        return meta

    def update_file_meta(self, file_id, meta):
        """Update metadata in redis cache

        Args:
          file_id (str): ID in file model
          meta (dict): keys to update
        """
        key = 'fid:%s:%s' % (self.uid, file_id[-8:])
        try:
            self.redis_conn.set(key, json.dumps(meta), ex=self.cache_ttl,
                                nx=True)
        except Exception as ex:
            logging.warning('action=update_file_meta uid=%s exception=%s' %
                            (self.uid, str(ex)))

    def del_file_meta(self, file_id, meta):
        """Remove a metadata item from cache

        Args:
          file_id (str): ID in file model
          meta (dict): keys to update
        """
        key = 'fid:%s:%s' % (self.uid, file_id[-8:])
        try:
            self.redis_conn.delete(key)
        except Exception as ex:
            logging.warning('action=del_file_meta uid=%s exception=%s' %
                            (self.uid, str(ex)))

    def upload_complete(self, file_id, status, func_worker):
        """Finalize upload - pass to the worker

        Args:
          file_id (str): ID of file meta in redis
          status (str): completion status from frontend ('done' if OK)
          func_worker (function): callback for passing to celery worker
        Returns:
          tuple: dict with file ID and message, and http status
        """

        meta = self.get_file_meta(file_id)
        if meta and status == 'done':
            duration = datetime.utcnow().timestamp() - meta['created']
            logging.info(dict(action='upload_complete', file_id=file_id,
                              duration='%.3f' % duration))
            func_worker(self.uid, file_id)
            Metrics().store('file_upload_bytes_total', labels=[
                'ctype=%s' % meta.get('ctype')], value=meta.get('size'))
            ret = True
            if meta.get('ctype') in Constants.MIME_IMAGE_TYPES:
                ret = Metrics(uid=self.uid, db_session=self.db).store(
                    'photo_daily_total')
            elif meta.get('ctype') in Constants.MIME_VIDEO_TYPES:
                ret = Metrics(uid=self.uid, db_session=self.db).store(
                    'video_daily_total')
            if not ret:
                # A user should not be able to get an upload URL after
                # reaching limit, but we log here just in case
                logging.warning(dict(
                    uid=self.uid, action='upload', size=meta.get('size'),
                    ctype=meta.get('ctype'), message='limit exceeded'))
            return dict(file_id=file_id), 201
        else:
            msg = _('unhandled frontend status')
            logging.warning('action=upload_complete status=%s message=%s' %
                            (status, msg))
            return dict(file_id=file_id, message=msg), 405

    def fetch_album_meta(self, album_id, thumbnail_height):
        """Retrieve metadata of a media album in format suitable
        for react-image-gallery

        Args:
          album_id: ID in album model
          thumbnail_height (int): which scaled image to choose
        Returns:
          dict: a set of image descriptors suitable for display
        """

        try:
            album = self.db.query(self.models.Album).filter(
                self.models.Album.id == album_id,
                self.models.Picture.status == 'active').one()
        except NoResultFound:
            return []
        results = []
        for picture in album.pictures:
            orig_uri = None
            uri_path = '%s/%s' % (picture.storage.cdn_uri, picture.path)
            fmt = picture.format_original
            sizes = [int(x) for x in album.sizes.split(',')
                     if not picture.height or int(x) <= picture.height]
            if fmt in Constants.MIME_VIDEO_TYPES:
                # TODO see line 249 of xiaolin's app.js example
                logging.warning('resource=album action=get id=%s '
                                'message=videos_nyi' % picture.id)
                continue
            else:
                max_height = Grants(self.db).get('photo_res_max')
                if not picture.height or picture.height <= max_height:
                    orig_uri = '%s.%s' % (uri_path, fmt)
                else:
                    orig_uri = '%s.%d.%s' % (uri_path, max_height, fmt)
            results.append(dict(
                id=picture.id,
                description=picture.caption,
                duration=picture.duration,
                height=picture.height,
                imageSet=[dict(srcSet='%s.%d.%s' % (uri_path, res, fmt),
                               media='(max-width: %dpx)' % res) for
                          res in sizes],
                name=picture.name,
                original=orig_uri,
                path=picture.path,
                size=picture.size,
                thumbnail='%s.%d.%s' % (uri_path, thumbnail_height, fmt),
                type='%s/%s' % ('video' if fmt in Constants.MIME_VIDEO_TYPES
                                else 'image', fmt),
                uid=picture.uid,
                width=picture.width))
            if picture.height and picture.height <= max_height:
                results[-1]['imageSet'].append(dict(
                    srcSet='%s.%s' % (uri_path, fmt),
                    media='(max-width: %dpx)' % picture.height))
                results[-1]['imageSet'].append(dict(
                    srcSet='%s.%s' % (uri_path, fmt),
                    media='(min-width: %dpx)' % picture.height))
            else:
                results[-1]['imageSet'].append(dict(
                    srcSet='%s.%d.%s' % (uri_path, sizes[-1], fmt),
                    media='(min-width: %dpx)' % sizes[-1]))
        return results

    def get_object(self, storage_id, storage_key):
        """Get object from bucket

        Args:
          storage_id (str): ID in storage model
          storage_key (str): pathname within storage bucket
        """
        try:
            storage = self.db.query(self.models.Storage).filter_by(
                id=storage_id, status='active').one()
        except NoResultFound:
            logging.error('storage volume not found')
        if storage.credentials.vendor == 'aws':
            return StorageS3(credentials=storage.credentials).get_object(
                                 storage.bucket, storage_key)

    def put_object(self, storage_id, storage_key, content,
                   content_type='image/jpeg'):
        """Send a file to the storage bucket

        Args:
          storage_id (str): ID in storage model
          storage_key (str): pathname within storage bucket
          content (bytes): file content
          content_type (str): mime type
        """
        try:
            storage = self.db.query(self.models.Storage).filter_by(
                id=storage_id, status='active').one()
        except NoResultFound:
            logging.error('storage volume not found')
        if storage.credentials.vendor == 'aws':
            return StorageS3(credentials=storage.credentials).put_object(
                                 storage.bucket, storage_key, content,
                                 content_type=content_type)

    def del_object(self, storage_id, storage_key):
        """Remove a file from a storage bucket

        Args:
          storage_id (str): ID in storage model
          storage_key (str): pathname within storage bucket
        """
        try:
            storage = self.db.query(self.models.Storage).filter_by(
                id=storage_id, status='active').one()
        except NoResultFound:
            logging.error('storage volume not found')
        if storage.credentials.vendor == 'aws':
            return StorageS3(credentials=storage.credentials).del_object(
                                 storage.bucket, storage_key)


class StorageBackblaze(object):
    """Vendor-specific storage interface: Backblaze B2.
    Not yet implemented

    Args:
      model (obj): a storage model
      db_session (obj): existing db session
    """
    def __init__(self, model, db_session=None):
        self.api = B2Api(InMemoryAccountInfo())
        self.db = db_session
        self.model = model

    def get_upload_url(self, storage_id, storage_key,
                       content_type='image/jpeg'):
        """Get a presigned authorization URL for upload

        Args:
          storage_id (str): ID of storage model object
          storage_key (str): pathname of new object to store
          content_type (str): mime type
        Returns:
          tuple: dict with upload URL, http status
        """
        try:
            storage = self.db.query(self.model).filter_by(
                id=storage_id, status='active').one()
        except NoResultFound:
            return dict(id=storage_id, message='storage volume not found'), 404

        self.api.authorize_account(
            'production', storage.credentials.key, storage.credentials.secret)
        try:
            result = self.api.raw_api.get_upload_url(
                self.api.account_info.get_api_url(),
                self.api.account_info.get_account_auth_token(),
                storage.identifier)
        except Exception as ex:
            msg = str(ex)
            return dict(message=msg, id=storage_id), 405
        return result, 201


class StorageS3(object):
    """Vendor-specific storage interface: Amazon S3

    Args:
      credential_ttl (int): how long until temp credential expires
      credentials (obj): API key and secret
      access_key (str): API key (if credentials object not specified)
      secret_key (str): API secret (if credentials object not specified)
      region (str): data center region specification
    """
    def __init__(self, credential_ttl=3600, credentials=None, access_key=None,
                 secret_key=None, region=Constants.DEFAULT_AWS_REGION):
        self.credential_ttl = credential_ttl
        self.region = region
        if credentials:
            self.access_key = credentials.key
            self.secret_key = credentials.secret
        else:
            self.access_key = access_key
            self.secret_key = secret_key
        self.api = boto3.client(
            's3', aws_access_key_id=self.access_key,
            aws_secret_access_key=self.secret_key,
            region_name=self.region,
            # signature_version = 's3v4',
            config=Config(s3={'addressing_style': 'virtual'}))

    def get_upload_url(self, bucket, storage_key,
                       content_type='image/jpeg'):
        """Get a presigned authorization URL for upload

        Args:
          bucket (str): name of bucket
          storage_key (str): pathname of new object to store
          content_type (str): mime type
        Returns:
          tuple: dict with uploadUrl, http status
        """
        try:
            result = self.api.generate_presigned_post(
                bucket, storage_key, ExpiresIn=self.credential_ttl)
        except ClientError as ex:
            msg = str(ex)
            return dict(message=msg, bucket=bucket), 405
        return dict(fields=result['fields'], uploadUrl=result['url']), 201

    def get_object(self, bucket, storage_key):
        """Fetch the object into a byte array

        Args:
          bucket (str): name of bucket
          storage_key (str): pathname of new object to store
        Returns:
          bytes
        """
        return self.api.get_object(Bucket=bucket,
                                   Key=storage_key)['Body'].read()

    def put_object(self, bucket, storage_key, content,
                   content_type='image/jpeg'):
        """Store a byte-array object

        Args:
          bucket (str): name of bucket
          storage_key (str): pathname of new object to store
          content (bytes): object
          content_type (str): mime type
        Returns:
          tuple: vendor results
        """
        logging.debug(dict(action='put', bucket=bucket,
                           storage_key=storage_key, content_type=content_type))
        return self.api.put_object(
            Bucket=bucket, Key=storage_key, Body=content,
            ContentType=content_type)

    def del_object(self, bucket, storage_key):
        """Delete an object

        Args:
          bucket (str): name of bucket
          storage_key (str): pathname of new object to store
        Returns:
          tuple: vendor results
        """
        logging.debug(dict(action='del', bucket=bucket,
                           storage_key=storage_key))
        return self.api.delete_object(Bucket=bucket, Key=storage_key)



"""profile controller

created 27-may-2019 by richb@instantlinux.net
"""

from apicrud import BasicCRUD


class ProfileController(BasicCRUD):
    def __init__(self):
        super().__init__(resource='profile')



"""session_manager.py

created 8-may-2019 by richb@instantlinux.net
"""

from datetime import datetime, timedelta
import json
import logging
import redis
import time

from .aes_encrypt import AESEncrypt
from .service_config import ServiceConfig
from .utils import gen_id

saved_redis = None


class SessionManager(object):
    """Session Manager - for active user sessions

    Each login session is stored as an encrypted JSON dict in redis, indexed
    by sub:token

    Args:
      ttl (int): seconds until a session expires
      redis_conn (obj): connection to redis service
    """
    def __init__(self, ttl=None, redis_conn=None):
        global saved_redis

        self.config = ServiceConfig().config
        self.connection = (
            redis_conn or saved_redis or redis.Redis(
                host=self.config.REDIS_HOST,
                port=self.config.REDIS_PORT, db=0))
        saved_redis = self.connection
        self.ttl = ttl or self.config.REDIS_TTL
        self.aes = AESEncrypt(self.config.REDIS_AES_SECRET)

    def create(self, uid, roles, key_id=None, **kwargs):
        """Create a session, which is an encrypted JSON object with the values
        defined in https://tools.ietf.org/html/rfc7519 for JWT claim names:

        - exp - expiration time, as integer Unix epoch time
        - iss - a constant JWT_ISSUER
        - jti - JWT ID, the randomly-generated token
        - sub - the uid of a user

        We add these:

        - auth - authorized roles
        - any other key=value pairs the caller passes as kwargs

        The session automatically expires based on object's ttl. Part of
        the jti token is used in redis key, to allow a user to log into
        multiple sessions. The rest of the token is encrypted, to
        secure it from replay attack in the event redis traffic
        is compromised.

        Args:
          uid: User ID
          roles: Authorized roles
          key_id: session key ID for redis (defaults to uid)
          nonce: a unique identifier for the token (random if not specified)
          ttl: duration of session (defaulted from class init)
        Returns:
          dict:
            Keys include auth (authorized roles), exp / iss / jti / sub
            (as above), along with parameters passed into this function
        """

        token = kwargs.pop('nonce', gen_id(prefix=''))
        ttl = kwargs.pop('ttl', self.ttl)
        params = dict(
            auth=':'.join(roles),
            exp=(datetime.utcnow() + timedelta(
                seconds=ttl)).strftime('%s'),
            iss=self.config.JWT_ISSUER, jti=token, sub=uid
        )
        key = 'ses:%s:%s' % (key_id or uid, token[-3:])
        content = {**params, **kwargs}  # noqa
        try:
            if self.connection.set(key, self.aes.encrypt(json.dumps(
                    content)), ex=ttl, nx=True):
                return content
        except redis.exceptions.ConnectionError as ex:
            logging.error(dict(action='session.create',
                               uid=uid, message=str(ex)))

    def get(self, uid, token, arg=None, key_id=None):
        """Get one or all key-value pairs stored by session create

        Args:
          uid (str): User ID
          token (str): The token value passed from create as 'jti'
          arg (str): key of desired value (None to fetch all)
          key_id (str): session key ID for redis (defaults to uid)
        Returns:
          dict or str: single value or dictionary of all session keys
        """
        key = 'ses:%s:%s' % (key_id or uid, token[-3:])
        try:
            data = json.loads(self.aes.decrypt(
                self.connection.get(key)))
        except TypeError:
            return None
        except Exception as ex:
            logging.info(dict(action='session.get', uid=uid,
                              exception=str(ex)))
            return None
        if data['jti'] != token:
            logging.warning(dict(action='session.get', message='rejected'))
            return None
        if arg:
            return data[arg] if arg in data else None
        return data

    def update(self, uid, token, arg, value, key_id=None):
        """Update a specified session key

        Args:
          uid: User ID
          token (str): The token value passed from create as 'jti'
          arg (str): key to update
          value (str): new value for key
          key_id (str): session key ID for redis (defaults to uid)
        """
        key = 'ses:%s:%s' % (key_id or uid, token[-3:])
        data = self.get(uid, token)
        data[arg] = value
        self.connection.set(key, self.aes.encrypt(
            json.dumps(data)), ex=self.ttl)

    def delete(self, uid, token, key_id=None):
        """Cancel a session

        Args:
          uid: User ID
          token (str): The token value passed from create as 'jti'
          key_id (str): session key ID for redis
        """
        try:
            self.connection.delete('ses:%s:%s' % (key_id or uid, token[-3:]))
        except redis.exceptions.ConnectionError as ex:
            logging.error(dict(action='session.delete', message=str(ex)))


class Mutex:
    """Simple mutex implementation for non-clustered Redis

    Args:
      lockname (str): a unique name for the lock
      redis_host (str): IP or DNS name of redis service
      maxwait (int): seconds to wait for a lock
      ttl (int): seconds to hold lock
      redis_conn (obj): existing redis connection
    """
    def __init__(self, lockname, redis_host=None, maxwait=20,
                 ttl=0, redis_conn=None):
        config = ServiceConfig().config
        self.connection = (
            redis_conn or redis.Redis(host=redis_host,
                                      port=config.REDIS_PORT, db=0))
        self.ttl = ttl or config.REDIS_TTL
        self.lock_signature = gen_id()
        self.lockname = lockname
        self.maxwait = maxwait

    def acquire(self):
        """Acquire a mutex lock

        Raises:
          TimeoutError: if the resource is unavailable
        """
        for retry in range(self.maxwait):
            try:
                if self.connection.set(
                        'mutex:%s' % self.lockname,
                        self.lock_signature, ex=self.ttl, nx=True):
                    return True
            except redis.exceptions.ConnectionError:
                pass
            time.sleep(1)
        raise TimeoutError('Unable to acquire mutex=%s' % self.lockname)

    def release(self):
        """Release a lock
        """
        if ((self.connection.get('mutex:%s' % self.lockname).decode('utf-8') !=
             self.lock_signature) or
                not self.connection.delete('mutex:%s' % self.lockname)):
            raise TimeoutError('Mutex held too long, may need a longer ttl')

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.release()



"""apikey.py

APIkey controller

created 27-dec-2020 by richb@instantlinux.net
"""

from apicrud import BasicCRUD, ServiceConfig
import models


class APIkeyController(BasicCRUD):
    """
    CRUD for API keys
    """
    def __init__(self):
        super().__init__(resource='apikey', model=models.APIkey)

    @staticmethod
    def create(body):
        return super(APIkeyController, APIkeyController).create(
            body, limit_related=dict(
                scopes=ServiceConfig().config.SCOPES_MAX))

    @staticmethod
    def update(id, body):
        body.pop('expires', None)
        return super(APIkeyController, APIkeyController).update(
            id, body, limit_related=dict(
                scopes=ServiceConfig().config.SCOPES_MAX))

    @staticmethod
    def delete(ids):
        """Undelete is disabled"""
        return super(
            APIkeyController, APIkeyController).delete(ids, force=True)



"""scope.py

Scopes - scoping restrictions

created 27-dec-2020 by richb@instantlinux.net
"""

from apicrud import BasicCRUD


class ScopeController(BasicCRUD):
    def __init__(self):
        super().__init__(resource='scope')



"""grants.py

created 27-may-2019 by richb@instantlinux.net
"""

from cachetools import TTLCache
from flask import g
import json
import logging

from .access import AccessControl
from .service_config import ServiceConfig
from . import utils


class Grants(object):
    """Account usage limits

    An account's usage limits are specified here in the grants table;
    the free-service tier is defined and passed in via load_defaults().
    Records in grants table are owned by administrator-level user.
    If a record matches a user uid, the default grant name=value is
    overridden.

    Attributes:
      db_session (obj): existing db session
      ttl (int): how long to cache a grant in memory
    """
    _singleton = None
    _cache = None

    def __init__(self, db_session=None, ttl=None):
        self.models = ServiceConfig().models
        self.session = db_session
        if not self.session:
            try:
                self.session = g.db
            except RuntimeError as ex:
                if 'Working outside of application context' not in str(ex):
                    raise
        config = ServiceConfig().config
        if 'defaults' not in self._cache:
            self._cache['defaults'] = config.DEFAULT_GRANTS

    def __new__(cls, db_session=None, ttl=None):
        if cls._singleton is None:
            cls._singleton = super(Grants, cls).__new__(cls)
            config = ServiceConfig().config
            cls._cache = TTLCache(maxsize=config.CACHE_SIZE,
                                  ttl=ttl or config.REDIS_TTL)
        return cls._singleton

    def get(self, name, uid=None):
        """Get the cached value of a named grant, if it hasn't expired

        Args:
          name (str): name of a grant, as defined in service config
          uid (str): user ID
        Returns:
          int or str: granted limit or None if undefined
        """
        if not uid:
            uid = AccessControl().uid
        if uid not in self._cache:
            records = self.session.query(self.models.Grant).filter_by(
                uid=uid, status='active').all()
            grants = {}
            for rec in records:
                if not rec.expires or rec.expires > utils.utcnow():
                    grants[rec.name] = rec.value
            self._cache[uid] = grants
        if name in self._cache[uid]:
            ret = self._cache[uid].get(name)
        else:
            if name not in self._cache['defaults']:
                logging.error('Grant name=%s undefined in config.yaml' % name)
            ret = self._cache['defaults'].get(name)
        try:
            return int(ret, 0)
        except (TypeError, ValueError):
            return ret

    def crud_get(self, crud_results, id):
        """Process results from BasicCRUD.get() for grants endpoint.
        If the id is found in database, perform the standard CRUD
        get(). Otherwise, look for a hybrid id in form uid:grant and
        return the cached Grant value. Grant values are serialized as
        strings even if they are integers (decimal, octal, hex).

        Args:
            crud_results (tuple): preliminary response
            name (str): name filter, if specified
        """
        if crud_results[1] == 200 or ':' not in id:
            return crud_results
        acc = AccessControl(model=self.models.Grant)
        uid, grantname = id.split(':')
        rbac = ''.join(sorted(list(acc.rbac_permissions(owner_uid=uid))))
        if 'r' not in rbac:
            return dict('access denied'), 403
        return dict(id=id,  uid=uid, name=grantname, value=str(self.get(
            grantname, uid=uid)), rbac=rbac, status='active'), 200

    def find(self, crud_results, **kwargs):
        """Process results from BasicCRUD.find() for grants endpoint

        Args:
            crud_results (tuple): preliminary response
            name (str): name filter, if specified
        """
        filter = json.loads(kwargs.get('filter', '{}'))
        name = kwargs.get('name') or filter.get('name')
        acc = AccessControl(model=self.models.Grant)
        uid = filter.get('uid') or acc.uid
        rbac = ''.join(sorted(list(acc.rbac_permissions(owner_uid=uid))))
        result = []
        for key, val in self._cache['defaults'].items():
            if name and key != name:
                continue
            for row in crud_results[0]['items']:
                if row['name'] == key:
                    result.append(row)
                    break
            else:
                result.append(dict(id='%s:%s' % (uid, key), uid=uid,
                                   name=key, value=str(val), rbac=rbac,
                                   status='active'))
        return dict(items=result, count=len(result)), crud_results[1]

    def uncache(self, uid):
        """Remove grants from cache, any time a user's status changes

        Args:
          uid (str): user ID
        """
        self._cache.pop(uid, None)

    def load_defaults(self, defaults):
        """Load default values from a dict of keyword: value pairs

        Args:
          defaults (dict): new defaults
        """
        self._cache['defaults'] = defaults



"""totp

created 5-apr-2021 by richb@instantlinux.net
"""
from flask import g
from flask_babel import _
import logging
from hashlib import md5
import pyotp
import random
import string
from urllib.parse import urlparse

from .. import AccessControl, AESEncryptBin, SessionAuth, database
from ..const import Constants


class AuthTOTP(SessionAuth):
    """TOTP for Session Authorization

    Functions for generating, registering and validating Time-based
    One Time Password tokens
    """
    def __init__(self):
        super().__init__()

    def generate(self):
        acc = AccessControl()
        logmsg = dict(action='totp_generate', identity=acc.identity)
        try:
            account = g.db.query(self.models.Account).filter_by(
                id=acc.account_id, status='active').one()
        except Exception as ex:
            return database.db_abort(dict(str(ex), **logmsg))
        if account.totp_secret:
            logging.info(dict(message='attempt to regenerate before revoke',
                              **logmsg))
            return dict(message=_(u'access denied')), 403
        totp = pyotp.random_base32()
        uri = pyotp.totp.TOTP(totp).provisioning_uri(
            name=account.owner.identity, issuer_name=self.config.APPNAME)
        return(dict(g.session.create(acc.uid, ['pendingtotp'],
                                     totp=totp, ttl=120, uri=uri))), 200

    def register(self, body):
        acc = AccessControl()
        config = self.config
        logmsg = dict(action='totp_register', identity=acc.identity)
        totp_secret = g.session.get(acc.uid, body.get('nonce'), arg='totp')
        if not totp_secret:
            msg = _(u'access denied')
            logging.warning(dict(message=msg, error='totp session expired',
                                 **logmsg))
            return dict(message=msg), 403
        try:
            account = g.db.query(self.models.Account).filter_by(
                id=acc.account_id, status='active').one()
        except Exception as ex:
            return database.db_abort(dict(str(ex), **logmsg))
        if account.totp_secret:
            logging.info(dict(message='attempt to register before revoke',
                              **logmsg))
            return dict(message=_(u'access denied')), 403
        if body.get('otp_first') != pyotp.TOTP(totp_secret).now():
            msg = _(u'access denied')
            logging.warning(dict(message=msg, error='wrong otp given',
                                 **logmsg))
            return dict(message=msg), 403
        backup_codes, hashed_backup_codes = [], []
        for i in range(config.LOGIN_MFA_BACKUP_CODES):
            chars = string.digits + string.ascii_lowercase
            code = ''.join(random.choice(chars) for x in range(
                Constants.MFA_BACKUP_CODELEN))
            backup_codes.append(code)
            hashed_backup_codes.append(md5(code.encode()).digest())
        account.totp_secret = totp_secret
        account.backup_codes = AESEncryptBin(
            config.DB_AES_SECRET).encrypt(b''.join(hashed_backup_codes))
        g.db.add(account)
        try:
            g.db.commit()
        except Exception as ex:
            return database.db_abort(str(ex), rollback=True, **logmsg)
        logging.info(dict(message=_(u'registered'), **logmsg))
        # Cancel any previous bypass cookie
        headers = {
            'Set-Cookie': '%s=; Domain=%s; Path=%s/auth; '
            'expires=Sat, Jan 01 2000 00:00:00 UTC; '
            'HttpOnly; SameSite=Strict; Secure' % (
                config.LOGIN_MFA_COOKIE_NAME,
                urlparse(config.PUBLIC_URL).hostname, config.BASE_URL)}
        return dict(message=_(u'registered'),
                    backup_codes=backup_codes), 200, headers



"""trashcan.py

Trashcan

created 22-apr-2021 by richb@instantlinux.net
"""
from flask_babel import _

from apicrud import BasicCRUD, state


class TrashcanController(BasicCRUD):
    def __init__(self):
        self.resource = 'trashcan'
        if self.resource not in state.controllers:
            state.controllers[self.resource] = self

    @staticmethod
    def find(**kwargs):
        """Find deleted resources

        Args:
            kwargs: as defined in openapi.yaml
        """
        return dict(count=0, items=[]), 200

    @staticmethod
    def get(id):
        """Get one deleted resource by ID

        Args:
            id (str): Database or hybrid grant ID
        """
        return dict(message=_(u'not found')), 404

    @staticmethod
    def update(id, body):
        """Update deleted resource - undelete
        """
        return dict(message=_(u'not found')), 404

    @staticmethod
    def delete(ids):
        """Remove resources

        Args:
          ids (list of str): record IDs to be flagged for removal
        Returns:
          tuple:
            first element is a dict with the id, second element is
            response code (200 on success)
        """
        return dict(message=_(u'not found')), 404



"""metrics.py

Metrics and usage tracking

created 21-feb-2021 by richb@instantlinux.net
"""

from flask_babel import _
import json
import logging
import os

from . import state
from .access import AccessControl
from .const import Constants
from .grants import Grants
from .service_config import ServiceConfig
from .service_registry import ServiceRegistry

INTERVALS = dict(hour=3600, day=3600 * 24, week=3600 * 24 * 7,
                 month=3600 * 24 * 31, indefinite=None)


class Metrics(object):
    """This implementation supports standard system metrics and two types of
    usage-billing: granted limits per period, or metered usage.

    See this article for a description of how redis makes the implementation
    straightforward:
    https://www.infoworld.com/article/3230455/how-to-use-redis-for-real-time-metering-applications.html
    The data store is in-memory, with snapshot/append persistence to disk.

    All metrics are defined in the metrics section of service_config.yaml.
    These follow a de facto standard described in best-practices
    documentation section of prometheus.io website. At time of implementation,
    flask prometheus client is not mature/user-friendly enough or
    suitable for usage-billing so to keep this consistent with
    service_config.yaml, there is no intent now or in the future to
    use it.

    Thanks to the expiring keys feature of redis, grant limits can be
    implemented without any periodic cleanup task. Upon first use,
    a decrementing key is created with expiration set to the configured period.
    If it counts down to zero before expiration (for example, 50 video uploads
    per day), the user is denied access to the resource until the key expires.
    If a user never returns, there will be no redis keys consuming system
    resources unless and until the user comes back, at which point a new grant
    period starts. Scaling to tens of millions of users is practical within a
    small footprint, and this is why user-tracking metrics are stored in redis
    rather than a traditional database.

    Attributes:
      uid (str): ID of a user, for usage tracking
      db_session (obj): existing db connection
      func_send (obj): function name for sending message via celery
    """
    def __init__(self, uid=None, db_session=None, func_send=None):
        self.config = config = ServiceConfig().config
        self.connection = state.redis_conn
        self.db_session = db_session
        self.func_send = state.func_send = func_send or state.func_send
        self.metrics = config.METRICS
        self.uid = uid

    def store(self, name, labels=[], value=None):
        """Store a metric: redis keys are in form mtr:<name>:<labels>
        To avoid possible ambiguity and multiple redis keys for the same
        label set, this function sorts labels before storing.

        Params:
          name (str): a metric name
          labels (list): labels, usually in form <label>=<value>
          value (int or float): value to store

        Returns:
          bool: true in all cases except for grant exceeded
        """
        if name not in self.metrics:
            raise AssertionError('name=%s not defined in config' % name)

        metric = self.metrics[name]
        if not labels:
            if metric['scope'] == 'instance':
                labels = ['instance=%s' % ServiceRegistry().get()['id']]
            else:
                labels = ['uid=%s' % self.uid] if self.uid else []
        labels.sort()
        key = 'mtr:%s:%s' % (name, ",".join(labels))
        if metric['style'] == 'grant':
            current = self.connection.get(key)
            limit = Grants(db_session=self.db_session).get(name, uid=self.uid)
            if current is None or int(current) >= limit:
                self.connection.set(key, limit - 1,
                                    ex=INTERVALS[metric['period']])
            elif int(current) <= 0:
                return False
            else:
                if metric['notify'] and self.func_send:
                    used = float(current) / limit * 100
                    used_next = (float(current) - 1.0) / limit * 100
                    if ((used >= 100 - metric['notify']) and
                            (used_next < 100 - metric['notify'])):
                        self.func_send(
                            to_uid=self.uid, template='usage_notify',
                            percent=metric['notify'], period=metric['period'],
                            resource=name)
                self.connection.decr(key)
        elif metric['style'] == 'counter':
            if type(value) is int:
                self.connection.incrby(key, value)
            elif type(value) is float:
                self.connection.incrbyfloat(key, value)
            elif value is None:
                self.connection.incr(key)
            else:
                raise AssertionError('Invalid value')
        elif metric['style'] == 'gauge':
            self.connection.set(key, value)
        return True

    def check(self, name):
        """Check remaining credit against grant-style metric

        Params:
          name (str): a metric name

        Returns:
          float: amount of credit remaining, or None
        """
        if name not in self.metrics:
            raise AssertionError('name=%s not defined in config' % name)

        metric = self.metrics[name]
        labels = ['uid=%s' % self.uid] if self.uid else []
        key = 'mtr:%s:%s' % (name, ",".join(labels))
        if metric['style'] != 'grant':
            raise AssertionError('name=%s is not a grant metric' % name)

        current = self.connection.get(key)
        limit = Grants(db_session=self.db_session).get(name, uid=self.uid)
        if current is None or float(current) >= limit:
            return float(limit)
        elif float(current) > 0:
            return float(current)
        return None

    def find(self, **kwargs):
        """Look up metrics defined by filter

        Returns:
          tuple (results dict, status)
        """
        try:
            filter = json.loads(kwargs.get('filter', '{}'))
        except json.decoder.JSONDecodeError as ex:
            logmsg = dict(action='find', resource='metric')
            msg = _(u'invalid filter specified') + '=%s' % str(ex)
            logging.error(dict(message=msg, **logmsg))
            return dict(message=msg), 405
        name = filter.get('name') or kwargs.get('name', '*')
        acc = AccessControl()
        auth = acc.auth
        if 'admin' in auth:
            if 'uid' in filter or 'uid' in kwargs:
                uid = filter.get('uid') or kwargs.get('uid')
                label = 'uid=%s' % uid if uid else 'uid=*'
            else:
                label = filter.get('label') or kwargs.get('label', '*')
                if label != '*':
                    label = '*%s*' % label
            self._process_collect()
        elif 'user' in auth:
            uid = filter.get('uid')
            if not uid or uid != acc.uid:
                return dict(message=_(u'access denied')), 403
            label = 'uid=%s' % uid
        else:
            return dict(message=_(u'access denied')), 403
        key = 'mtr:%s:%s' % (name, label)
        limit = kwargs.get('limit', Constants.PER_PAGE_DEFAULT)
        offset = kwargs.get('offset', 0)
        count, retval = 0, []
        for item in self.connection.scan_iter(match=key):
            if offset:
                offset -= 1
            else:
                value = self.connection.get(item).decode()
                try:
                    value = int(value)
                except ValueError:
                    value = float(value)
                retval.append(dict(
                    name=item.decode().split(':')[1],
                    labels=item.decode().split(':')[2].split(','),
                    value=value))
            count += 1
            if count >= limit:
                break
        return dict(items=retval, count=count), 200

    def collect(self, **kwargs):
        """
        Prometheus-compatible metrics exporter. Collects current
        metrics from redis, evaluates metric-type settings in
        service_config.yaml, and returns a plain-text result in this form:

        # TYPE process_resident_memory_bytes gauge
        process_resident_memory_bytes{instance="k8s-01.ci.net"} 20576.0
        # TYPE process_start_time_seconds gauge
        process_start_time_seconds{instance="k8s-01.ci.net"} 1614051156.14
        # TYPE api_calls_total counter
        api_calls_total{resource="person"} 4
        """
        try:
            filter = json.loads(kwargs.get('filter', '{}'))
        except json.decoder.JSONDecodeError as ex:
            logmsg = dict(action='find', resource='metric')
            msg = _(u'invalid filter specified') + '=%s' % str(ex)
            logging.error(dict(message=msg, **logmsg))
            return dict(message=msg), 405
        name = filter.get('name') or kwargs.get('name', '*')
        if 'uid' in filter:
            uid = filter.get('uid', '*')
            label = 'uid=%s' % uid if uid else 'uid=*'
        else:
            label = filter.get('label') or kwargs.get('label', '*')
            if label != '*':
                label = '*%s*' % label
        self._process_collect()
        key = 'mtr:%s:%s' % (name, label)
        limit = kwargs.get('limit', Constants.PER_PAGE_DEFAULT)
        count, retval = 0, []
        for item in self.connection.scan_iter(match=key):
            name = item.decode().split(':')[1]
            if name not in self.metrics:
                continue
            type_ = self.metrics[name]['style']
            if type_ == 'grant':
                # TODO decide better way to filter out style=grant
                if 'uid' not in label:
                    continue
                type_ = 'gauge'
            retval.append(dict(name=name, type_=type_,
                               labels=item.decode().split(':')[2].split(','),
                               value=self.connection.get(item).decode()))
            count += 1
            if count >= limit:
                break
        output = ""
        for item in retval:
            output += "# TYPE %s %s\n" % (item['name'], item['type_'])
            labels = ['%s="%s"' % (label.split('=')[0], label.split('=')[1])
                      if '=' in label else label
                      for label in item['labels']]
            output += "%s%s %s\n" % (
                item['name'],
                '{%s}' % ','.join(labels) if labels else '', item['value'])
        return output, 200

    def _process_collect(self):
        """Collect metrics for process_xxx"""

        proc = '/proc/self'
        boot_timestamp = self._boot_time()
        if not boot_timestamp:
            return None
        with open('%s/stat' % proc, 'rb') as stat:
            # https://man7.org/linux/man-pages/man5/procfs.5.html
            #  subtract 3 from index values documented in /proc/[pid]/stat
            #  we use utime, stime, starttime, vsize and rss
            stats = (stat.read().split(b')')[-1].split())
        try:
            with open('%s/limits' % proc, 'rb') as fp:
                for line in fp:
                    if line.startswith(b'Max open files'):
                        max_fds = int(line.split()[3])
        except (IOError, OSError):
            max_fds = 0
        try:
            ticks = os.sysconf('SC_CLK_TCK')
        except (ValueError, TypeError, AttributeError, OSError):
            ticks = 100.0

        self.store('process_cpu_seconds_total',
                   value=(float(stats[11]) + float(stats[12])) / ticks)
        self.store('process_max_fds', value=max_fds)
        self.store('process_open_fds', value=len(os.listdir('%s/fd' % proc)))
        self.store('process_resident_memory_bytes', value=float(stats[21]))
        self.store('process_start_time_seconds',
                   value=boot_timestamp + float(stats[19]) / ticks)
        self.store('process_virtual_memory_bytes', value=float(stats[20]))

    @staticmethod
    def _boot_time():
        try:
            with open('/proc/stat', 'rb') as fp:
                for line in fp:
                    if line.startswith(b'btime '):
                        return float(line.split()[1])
        except IOError:
            return None



"""health.py

created 23-sep-2019 by richb@instantlinux.net
"""
from apicrud import health, ServiceConfig

from example import _version
from models import AlembicVersion


class HealthController(object):
    """Healthcheck endpoint - return a standardized status check"""

    @staticmethod
    def get(tests=None):
        """Pass application-specific parameters to the healthcheck
        function

        Returns:
          tuple:
            first element is a dict containing health check report,
            with optional tests, in format recommended by
            Nadareishvili in RFC draft
            https://tools.ietf.org/id/draft-inadarei-api-health-check-04.html;
            second element is http response code
        """
        config = ServiceConfig().config
        return health.healthcheck(
            app_name=config.APPNAME, service_name=config.SERVICE_NAME,
            tests=tests, model=AlembicVersion,
            releaseId=_version.vcs_ref, build_date=_version.build_date,
            version=_version.__version__)



"""local_func

created 26-mar-2020 by richb@instantlinux.net
"""
from datetime import datetime, timedelta
from flask import g
from flask_babel import _
import logging
from passlib.hash import sha256_crypt
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy_utils.types.encrypted.padding import InvalidPaddingError
import time

from .. import Metrics, ServiceConfig
from ..database import db_abort


def login(username, password):
    """Login using credentials stored in local database.  If
    successful, this passes a sqlalchemy account record back for
    further processing by SessionAuth._login_accepted.

    Args:
      username (str): the username or email identity
      password (str): password

    Returns:
      tuple: dict with error message, http status (200 if OK), account object
    """
    config = ServiceConfig().config
    models = ServiceConfig().models
    try:
        if '@' in username:
            account = g.db.query(models.Account).join(
                models.Person).filter(
                    models.Person.identity == username,
                    models.Account.status == 'active').one()
            username = account.name
        else:
            account = g.db.query(models.Account).filter_by(
                name=username, status='active').one()
    except InvalidPaddingError:
        logging.error('action=login message="invalid DB_AES_SECRET"')
        return dict(message=_(u'DB operational error')), 503
    except (NoResultFound, TypeError):
        return dict(username=username, message=_(u'not valid')), 403
    except OperationalError as ex:
        return db_abort(str(ex), action='login')
    if (account.invalid_attempts >= config.LOGIN_ATTEMPTS_MAX and
        account.last_invalid_attempt + timedelta(
            seconds=config.LOGIN_LOCKOUT_INTERVAL) > datetime.utcnow()):
        Metrics().store('logins_fail_total')
        time.sleep(5)
        msg = _(u'locked out')
        logging.warning(dict(message=msg, identity=account.owner.identity))
        return dict(username=username, message=msg), 403
    if account.password == '':
        logging.error("username=%s, message='no password'" % username)
        return dict(username=username, message=_(u'no password')), 403
    elif not password or not sha256_crypt.verify(password, account.password):
        account.invalid_attempts += 1
        account.last_invalid_attempt = datetime.utcnow()
        Metrics().store('logins_fail_total')
        logging.info(
            'action=login username=%s credential=invalid attempt=%d' %
            (username, account.invalid_attempts))
        g.db.commit()
        return dict(username=username, message=_(u'not valid')), 403

    return {}, 200, account



"""initialize.py

created 9-jan-2021 by richb@instantlinux.net
"""
from authlib.integrations.flask_client import OAuth
import connexion
from datetime import datetime
from flask import abort, g, jsonify, request
from flask_cors import CORS
import logging
import os
import redis

from . import database, AccessControl, AccountSettings, Metrics, RateLimit, \
    ServiceConfig, ServiceRegistry, state
from .const import Constants
from .session_manager import SessionManager


def app(application, controllers, models, path, redis_conn=None,
        db_url=None, db_seed_file=None, func_send=None):
    """Initialize the Flask app defined by openapi.yaml: first four
    params must be passed from main; optional params here support
    unit-test framework.

    Args:
      application (obj): a connexion object
      controllers (obj): all controllers
      models (obj): all models
      path (str): location of configuration .yaml / i18n files
      db_seed_file (filename): database records in yaml format
      db_url (str): URL of database
      func_send (obj): application's function to send messages
      redis_conn (obj): connection to redis

    Returns:
      obj: Flask app
    """
    kwargs = {}
    if db_seed_file:
        kwargs['db_seed_file'] = db_seed_file
    config = ServiceConfig(
        babel_translation_directories='i18n;%s' % os.path.join(path, 'i18n'),
        file=os.path.join(path, 'config.yaml'),
        models=models, reset=True, **kwargs).config
    db_url = db_url or config.DB_URL
    logging.basicConfig(level=config.LOG_LEVEL,
                        format='%(asctime)s %(levelname)s %(message)s',
                        datefmt='%m-%d %H:%M:%S')
    logging.getLogger('werkzeug').setLevel(logging.ERROR)
    application.app.config.from_object(config)
    application.add_api(
        os.path.join(os.path.dirname(os.path.abspath(__file__)),
                     Constants.SERVICE_CONFIG_FILE), base_path='/config/v1')
    application.add_api(config.OPENAPI_FILE)
    application.add_error_handler(400, render_status_4xx)
    application.add_error_handler(429, render_status_4xx)
    application.add_error_handler(connexion.ProblemException,
                                  render_problem)
    CORS(application.app,
         resources={r"/api/*": {'origins': config.CORS_ORIGINS}},
         supports_credentials=True)
    state.config = config
    state.func_send = func_send
    state.models = models
    state.oauth['init'] = OAuth(application.app)
    state.redis_conn = redis_conn or redis.Redis(
        host=config.REDIS_HOST, port=config.REDIS_PORT, db=0)
    ServiceRegistry().register(controllers.resources())
    if database.initialize_db(db_url=db_url, redis_conn=redis_conn):
        Metrics().store(
            'api_start_timestamp', value=int(datetime.now().timestamp()))
    AccessControl().load_rbac(config.RBAC_FILE)
    for provider in config.AUTH_PARAMS.keys():
        client_id = os.environ.get('%s_CLIENT_ID' % provider.upper())
        client_secret = os.environ.get('%s_CLIENT_SECRET' % provider.upper())
        if client_id:
            state.oauth['init'].register(name=provider, client_id=client_id,
                                         client_secret=client_secret,
                                         **config.AUTH_PARAMS[provider])
            logging.info(dict(action='initialize', provider=provider))

    logging.info(dict(action='initialize_app', port=config.APP_PORT))
    return application.app


def before_request():
    """flask session setup - database and metrics"""
    g.db = database.get_session()
    g.session = SessionManager()
    g.request_start_time = datetime.utcnow()
    try:
        resource = request.url_rule.rule.split('/')[3]
    except Exception:
        resource = None
    if resource != 'metrics':
        Metrics().store('api_calls_total', labels=['resource=%s' % resource])
    if request.method != 'OPTIONS' and RateLimit().call():
        Metrics().store('api_errors_total', labels=['code=%d' % 429])
        abort(429)


def after_request(response):
    """flask headers and metrics - all responses get a cache-control header"""
    config = ServiceConfig().config
    response.cache_control.max_age = config.HTTP_RESPONSE_CACHE_MAX_AGE
    if config.AUTH_SKIP_CORS:
        try:
            if request.url_rule.rule.split('/')[3] == 'auth':
                response.headers['Access-Control-Allow-Origin'] = '*'
                response.headers['Access-Control-Allow-Headers'] = (
                    'Content-Type')
        except (AttributeError, IndexError):
            pass
    Metrics().store(
        'api_request_seconds_total', value=datetime.utcnow().timestamp() -
        g.request_start_time.timestamp())
    return response


def get_locale():
    """flask locale"""
    acc = AccessControl()
    if acc.auth and acc.uid:
        locale = AccountSettings(acc.account_id,
                                 uid=acc.uid, db_session=g.db).locale
        if locale:
            return locale
    return request.accept_languages.best_match(
        ServiceConfig().config.LANGUAGES)


def render_status_4xx(error):
    """Render function to provide message in dict as required for
    react-admin to display text of message for 4xx error codes

    Args:
      error (obj): the error object with name and description
    """
    Metrics().store('api_errors_total', labels=['code=%s' % error.code])
    return jsonify(dict(
        message=error.description,
        error=dict(status=error.name, code=error.code))), error.code


def render_problem(error):
    """Render function to provide message in dict as required for
    react-admin to display text of message for exception
    connexion.ProblemException

    Args:
      error (obj): the error object with name and description
    """
    Metrics().store('api_errors_total', labels=['code=%s' % error.status])
    return jsonify(dict(
        message=error.detail,
        error=dict(status=error.title, code=error.status))), error.status



"""database.py

Database session management and schema update

created 31-mar-2019 by richb@instantlinux.net
"""

import alembic.config
import alembic.script
from alembic.runtime.environment import EnvironmentContext
from flask import _app_ctx_stack, g
from flask_babel import _
import logging
import os.path
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.event import listen
from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.sql import select, func
import sys
import time
import yaml

from .const import Constants
from .session_manager import Mutex
from .service_config import ServiceConfig

Base = declarative_base()
db_engine = None
Session = None
spatialite_loaded = False


def get_session(scopefunc=None, scoped=True, db_url=None, engine=None):
    """open a db session scoped to flask context or celery thread

    Args:
      scopefunc (function): function which returns a unique thread ID
      scoped (bool): whether to use scoped session management
      db_url (str): URL of database
      engine (obj): override engine object (for unit tests)
    Returns:
      obj: session
    """

    global db_engine
    if not engine:
        # This is where init_db is invoked under celery
        engine = db_engine or _init_db(db_url=db_url)
    if scoped:
        global Session
        if scopefunc is None:
            scopefunc = _app_ctx_stack.__ident_func__
        Session = scoped_session(sessionmaker(autocommit=False,
                                              autoflush=False,
                                              bind=engine),
                                 scopefunc=scopefunc)
        if engine.url.drivername == 'sqlite':
            Session.execute('PRAGMA foreign_keys=on')
        return Session
    else:
        Session = sessionmaker(autocommit=False, autoflush=False, bind=engine)
        return Session


def initialize_db(db_url=None, engine=None, redis_conn=None):
    """initialize database connectivity

    Args:
      db_url (str): URL of database
      engine (obj): override engine object (for unit tests)
      redis_conn (obj): redis connection object override (for unit tests)

    Returns:
      bool: True if successful
    """
    config = ServiceConfig().config
    models = ServiceConfig().models
    _init_db(db_url=db_url, geo_support=config.DB_GEO_SUPPORT,
             engine=engine, connection_timeout=config.DB_CONNECTION_TIMEOUT)
    if config.DB_MIGRATE_ENABLE:
        try:
            with Mutex('alembic', redis_host=config.REDIS_HOST,
                       ttl=config.DB_SCHEMA_MAXTIME, redis_conn=redis_conn):
                schema_update(db_engine, models)
        except TimeoutError:
            logging.error(dict(action='init_db', message='mutex timeout'))
            return False
    else:
        # just wait for schema update
        schema_update(db_engine, models)
    return True


def db_abort(error, rollback=False, **kwargs):
    """Helper for logging exceptions
    Args:
      error (str): exception string for log
      rollback (bool): whether to roll back
      kwargs: any other values to log

    Returns:
      tuple: dict with generic message, status 500
    """
    msg = _(u'DB operational error')
    logging.error(dict(message=msg, error=error, **kwargs))
    g.db.rollback()
    return dict(message=msg), 500


def alembic_migrate(models, version, script_location, migrate=False,
                    db_session=None, schema_maxtime=0, seed_func=None):
    """run schema migrations

    Args:
      models (obj): the models file object
      version (str): schema version expected after migration
      script_location (str): relative path name of alembic's env.py
      migrate (bool): whether to run alembic migrations
      db_session (obj): existing db session
      schema_maxtime (int): how long to wait for migration
      seed_func (function): function to seed initial records in blank db
    """

    cfg = alembic.config.Config()
    cfg.set_main_option('script_location', script_location)
    script = alembic.script.ScriptDirectory.from_config(cfg)
    env = EnvironmentContext(cfg, script)
    if (version == script.get_heads()[0]):
        logging.info('action=schema_update version=%s is current' %
                     version)
    elif migrate:
        def _do_upgrade(revision, context):
            return script._upgrade_revs(script.get_heads(), revision)

        conn = None
        while schema_maxtime > 0:
            try:
                conn = db_engine.connect()
                break
            except OperationalError as ex:
                logging.info('action=alembic_migrate status=waiting message=%s'
                             % str(ex))
            schema_maxtime -= 10
            time.sleep(10)
        if not conn:
            logging.error('action=alembic_migrate status=timeout')
            sys.exit(1)
        if db_engine.dialect.name == 'sqlite' and spatialite_loaded:
            conn.execute(select([func.InitSpatialMetaData(1)]))
        env.configure(connection=conn, target_metadata=Base.metadata,
                      verbose=True, fn=_do_upgrade)
        with env.begin_transaction():
            env.run_migrations()
        logging.info('action=schema_update finished migration, '
                     'version=%s' % script.get_heads()[0])
    else:
        # Not migrating: must wait
        wait_time = schema_maxtime
        while version != script.get_heads()[0] and wait_time:
            time.sleep(5)
            wait_time -= 5
    if version is None and seed_func:
        if not db_session:
            db_session = get_session(scoped=True)
        seed_func(db_session)
        db_session.close()


def schema_update(db_engine, models):
    """Run alembic migrations for updating schema

    Must be called with mutex for duration of update to prevent race
    conditions; begin_transaction() does not ensure mutual exclusion.

    Args:
      db_engine (obj): connection to database
      models (obj): models to generate or update
    """
    db_session = get_session(scoped=True)
    config = ServiceConfig().config
    script_location = config.DB_MIGRATIONS
    try:
        version = db_session.query(models.AlembicVersion).one().version_num
    except (NoResultFound, OperationalError, ProgrammingError) as ex:
        logging.warning('DB schema does not yet exist: %s' % str(ex))
        version = None
    db_session.close()
    if config.DB_SCHEMA_MAXTIME == 0:
        logging.info('found schema version=%s, skipping update' % version)
    else:
        alembic_migrate(models, version, script_location,
                        migrate=config.DB_MIGRATE_ENABLE,
                        schema_maxtime=config.DB_SCHEMA_MAXTIME,
                        seed_func=seed_new_db)


def seed_new_db(db_session):
    """Seed a new db with admin account and related records. Reads
    records defined in a file db_seed.yaml (or as otherwise specified
    in the main config.yaml) and places them in the database. Timezone
    name records are loaded from this source directory's timezones.yaml
    unless at least one tz record is provided in db_seed.yaml.

    Args:
      db_session (obj): existing db session
    """

    if db_session.bind.dialect.name == 'sqlite':
        cmd = 'pragma foreign_keys'
    else:
        cmd = 'SET FOREIGN_KEY_CHECKS'
    db_session.execute('%s=off' % cmd)
    with open(ServiceConfig().config.DB_SEED_FILE, 'rt', encoding='utf8') as f:
        records = yaml.safe_load(f)
    if 'tz' not in records:
        with open(os.path.join(os.path.dirname(
                __file__), 'timezones.yaml'), 'rt', encoding='utf8') as f:
            records['tz'] = yaml.safe_load(f)['tz']
    models = ServiceConfig().models
    for resource, records in records.items():
        if resource == '_constants':
            continue
        for record in records:
            if 'geolat' in record:
                # Store lat/long as fixed-precision integers
                record['geolat'] *= 1e7
                record['geolong'] *= 1e7
            db_session.add(getattr(models, resource.capitalize())(**record))
    db_session.commit()
    db_session.execute('%s=on' % cmd)


def _init_db(db_url=None, engine=None, connection_timeout=0,
             geo_support=True):
    global db_engine
    if not db_engine:
        try:
            db_engine = engine or create_engine(
                db_url, pool_pre_ping=True,
                pool_recycle=connection_timeout)
        except Exception as ex:
            logging.error('action=init_db status=error message=%s' % str(ex))
            return None
    if db_engine.url.drivername == 'sqlite':
        db_engine.execute('PRAGMA foreign_keys=on')
        if geo_support:
            listen(db_engine, 'connect', _load_spatialite)
    db = get_session(scoped=True)
    if hasattr(db, 'query_property'):
        Base.query = db.query_property()
        db.remove()
    connection_timeout = 20
    while connection_timeout > 0:
        try:
            Base.metadata.create_all(bind=db_engine)
            break
        except OperationalError as ex:
            logging.info('action=init_db status=waiting message=%s' % str(ex))
        connection_timeout -= 10
        time.sleep(10)
    return db_engine


def _load_spatialite(dbapi_conn, connection_record):
    global spatialite_loaded
    for lib in Constants.LIB_MOD_SPATIALITE:
        if os.path.exists(lib):
            dbapi_conn.enable_load_extension(True)
            dbapi_conn.load_extension(lib)
            logging.info('action=load_spatialite message=geo_support_added')
            spatialite_loaded = True
            return



"""utils.py

Utilities
  Miscellaneous utility functions that don't fit elsewhere

created 11-apr-2020 by richb@instantlinux.net
"""
from datetime import datetime
from flask import g
from html.parser import HTMLParser
import random
import re
import string


def gen_id(length=8, prefix='x-', chars=(
        '-' + string.digits + string.ascii_uppercase + '_' +
        string.ascii_lowercase)):
    """Record IDs are random 48-bit RFC-4648 radix-64 with a fixed prefix
    to make them (somewhat) more human-recognizable

    First 15 bits are generated from Unix epoch to make them sortable
    by date (granularity 24 hours); rolls over after year 2107

    Args:
      length (int): length of generated portion after prefix
      prefix (str): prefix to distinguish ID type
      chars (str): set of characters to choose from for random portion
    """
    def _int2base(x, chars, base=64):
        return _int2base(x // base, chars, base).lstrip(chars[0]) + chars[
            x % base] if x else chars[0]
    return (prefix +
            _int2base((utcnow() - datetime(2018, 1, 1)).days * 8 +
                      random.randint(0, 8), chars, base=len(chars)) +
            ''.join(random.choice(chars) for i in range(length - 3)))


def req_duration():
    """Report request duration as milliseconds
    """
    return '%.3f' % (utcnow().timestamp() -
                     g.request_start_time.timestamp())


def utcnow():
    """For mocking: unittest.mock can't patch out datetime.utcnow directly """
    return datetime.utcnow()


def identity_normalize(identity):
    """Normalize an email address for use as an identity: downcase
    and in certain cases remove characters. This is required to secure
    against a type of attack involving password-resets: an example
    of the vulnerability is described here:
     https://jameshfisher.com/2018/04/07/the-dots-do-matter-how-to-scam-a-gmail-user/

    Args:
      identity (str): a raw email address
    Returns: str
    """
    if not identity:
        return
    if re.match('.*@gmail.com$', identity, re.I):
        identity = '%s@gmail.com' % identity.split('@')[0].replace('.', '')
    try:
        user, domain = identity.split('@')
        return ('%s@%s' % (user.split('+')[0], domain)).lower()
    except ValueError:
        return identity.lower()


def strip_tags(html):
    """Convert html to plain-text by stripping tags

    Args:
      html (str): an html document
    """
    s = HtmlStripper()
    s.feed(html)
    return s.get_data()


class HtmlStripper(HTMLParser):
    def __init__(self):
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.fed = []

    def handle_data(self, d):
        """ add a string to the html object

        Args:
          d (str): chunk of html
        """
        self.fed.append(d)

    def get_data(self):
        """ return the stripped html

        Returns: str
        """
        return ' '.join(self.fed)



"""oauth2_func

created 29-mar-2021 by richb@instantlinux.net
"""
from flask_babel import _
import logging
from urllib.parse import parse_qs, urlparse

from .. import ServiceConfig


def login(oauth_reg, method, cache=None):
    """Initiate OAuth2 login

    Args:
      oauth_reg (obj): pre-registered oauth object (from initialize.py)
      method (str): provider name
    """
    logmsg = dict(action='account_login', method=method)
    try:
        client = oauth_reg.create_client(method)
    except RuntimeError as ex:
        msg = _(u'login client missing')
        logging.error(dict(message=msg, error=str(ex), **logmsg))
        return dict(message=msg), 405
    config = ServiceConfig().config
    msg, error = _(u'openid client failure'), ''
    try:
        retval = client.authorize_redirect(
            '%s%s/%s/%s' % (config.PUBLIC_URL,
                            config.BASE_URL,
                            'auth_callback', method))
        if config.AUTH_SKIP_CORS:
            state = parse_qs(urlparse(retval.location).query).get(
                'state')[0]
            if cache:
                cache.set('session_foo_state', state, expires=120)
        if retval.status_code == 302:
            return dict(location=retval.location), 200
    except RuntimeError as ex:
        error = str(ex)
    logging.error(dict(message=msg, error=error, **logmsg))
    return dict(message=msg), 405



"""service_config.py

created 12-aug-2020 by docker@instantlinux.net
"""
import binascii
from collections import namedtuple
from flask_babel import gettext as _
import jsonschema
import logging
import os
import yaml

from .const import Constants

state = {}
config = None


class ServiceConfig(object):
    """Service configration - have it your way!

    Flask and application configuration global default values are defined in
    service_config.yaml here in the source directory. Overrides of these
    values can be specified in a few ways and are evaluated in this order:

    - Environment variables set by parent process
    - Values defined in a file in yaml format
    - Values passed as a keyword arg at first class invocation

    For runtime security, the config singleton is stored as an immutable
    namedtuple: to change values, update settings and restart the
    container running the service.

    An endpoint /config/v1/config provides read-only access to these values
    except those of type password.  Always override those secret values
    before deploying your service.

    Attribute keys specified as env vars are UPPERCASE, and attribute keys
    stored in the config object are also uppercase. Use lowercase to
    specify attribute keys in kwargs or the yaml input file.

    Args:
      file (str): path of a YAML file defining override values
      models (obj): sqlalchemy db models
      reset (boolean): reset cached values (for unit tests)
      **kwargs: key=value pair arguments to override values

    Raises:
      AttributeError if invalid specification
    """

    def __init__(self, file=None, models=None, reset=False, **kwargs):
        global config

        if reset or not config:
            with open(os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                   Constants.SERVICE_CONFIG_FILE), 'rt',
                      encoding='utf8') as f:
                openapi = yaml.safe_load(f)
            overrides = {}
            if file:
                with open(file, 'rt', encoding='utf8') as f:
                    overrides = yaml.safe_load(f)
                if not overrides:
                    raise AttributeError('No values found in %s' % file)

            for key, schema in openapi['components']['schemas'][
                    'Config']['properties'].items():
                if os.environ.get(key.upper()):
                    if schema['type'] == 'array':
                        state[key] = os.environ[key.upper()].split(',')
                    elif schema['type'] == 'integer':
                        state[key] = int(os.environ[key.upper()])
                    elif schema['type'] == 'boolean':
                        if os.environ[
                                key.upper()].lower() in ('1', 'true', 'yes'):
                            state[key] = True
                        elif os.environ[
                                key.upper()].lower() in ('0', 'false', 'no'):
                            state[key] = False
                        else:
                            raise AttributeError('Invalid value for %s' % key)
                    else:
                        state[key] = os.environ[key.upper()]
                elif key in overrides:
                    if key in ('metrics',):
                        state[key].update(overrides[key])
                    else:
                        state[key] = overrides[key]
                elif key in kwargs:
                    state[key] = kwargs[key]
                elif 'default' in schema:
                    state[key] = schema['default']
                elif key == 'db_url':
                    state[key] = self._compose_db_url()
                else:
                    raise AttributeError('No value specified for %s' % key)
            jsonschema.validate(instance=state, schema=openapi['components'][
                'schemas']['Config'])
            # Special cases
            state['secret_key'] = binascii.unhexlify(
                state['flask_secret_key'])
            state['metrics'] = self._compose_metrics(state['metrics'])
            state['log_level'] = self._compose_loglevel(state['log_level'])
            state['template_folders'].append(
                    os.path.join(os.path.dirname(__file__), 'templates'))
            if os.environ.get('APP_ENV') != 'prod':
                state['login_admin_limit'] = max(
                    state['login_admin_limit'], 86400)
                state['login_session_limit'] = max(
                    state['login_admin_limit'], 86400)
            if '/' not in state['babel_translation_directories'].split(';')[0]:
                state['babel_translation_directories'] = (
                    os.path.join(os.path.dirname(__file__),
                                 state['babel_translation_directories']))
            if file:
                if '/' not in state['rbac_file']:
                    state['rbac_file'] = os.path.join(os.path.dirname(
                        os.path.abspath(file)), state['rbac_file'])
                if '/' not in state['db_migrations']:
                    state['db_migrations'] = os.path.join(os.path.dirname(
                        os.path.abspath(file)), state['db_migrations'])
                if '/' not in state['db_seed_file']:
                    state['db_seed_file'] = os.path.join(os.path.dirname(
                        os.path.abspath(file)), state['db_seed_file'])
            state.pop('schema', None)
            state.pop('models', None)
            config = namedtuple('Struct', [
                key.upper() for key in state.keys()])(*state.values())
            state['schema'] = openapi['components']['schemas']['Config']

        if models:
            state['models'] = models
        self.models = state.get('models')
        self.config = config

    @staticmethod
    def _compose_db_url():
        return (
            '%(dbtype)s://%(user)s:%(password)s@%(endpoint)s/%(database)s' %
            {'dbtype': os.environ.get('DB_TYPE', 'mysql+pymysql'),
             'user': os.environ.get('DB_USER', 'example'),
             'password': os.environ.get('DB_PASS', 'example'),
             'endpoint': '%s:%d' % (os.environ.get('DB_HOST'),
                                    int(os.environ.get('DB_PORT', 3306))),
             'database': os.environ.get('DB_NAME', 'example')})

    @staticmethod
    def _compose_loglevel(level):
        """Convert a string 'info' to the enum logging.INFO

        Args:
            level (str): debug, info, warning, error or critical
        """
        return dict(
            debug=logging.DEBUG, info=logging.INFO, warning=logging.WARNING,
            error=logging.ERROR, critical=logging.CRITICAL)[level]

    @staticmethod
    def _compose_metrics(metrics):
        """Fill in default values for scope, style, period

        Args:
            metrics (dict): usage metric definitions
                period: hour, day, week, month, indefinite
                scope: user, instance, sitewide
                style: grant, usage
        """
        ret = metrics
        for key, item in metrics.items():
            if set(item.keys()) - set(['notify', 'period', 'scope', 'style']):
                raise AttributeError('Invalid metric configuration')
            ret[key]['notify'] = item.get('notify', 0)
            ret[key]['period'] = item.get('period', 'day')
            ret[key]['scope'] = item.get('scope', 'user')
            ret[key]['style'] = item.get('style', 'grant')
        return ret

    def set(self, key, value):
        """Set a single value

        Args:
            key (str) - attribute name
            value - new value
        """
        global config

        jsonschema.validate(instance={key: value}, schema=state['schema'])
        state[key] = value
        config = namedtuple('Struct', [key.upper() for key in
                                       state.keys()])(*state.values())

    @staticmethod
    def get():
        from .access import AccessControl

        if not AccessControl().auth or 'admin' not in AccessControl().auth:
            return dict(message=_(u'access denied')), 403
        retval = {key: state[key]
                  for key, props in state['schema']['properties'].items()
                  if key not in ('db_url') and
                  ('format' not in props or props['format'] != 'password')}
        return retval, 200
